#!/usr/bin/env python

#from twisted.internet import gtk2reactor
#gtk2reactor.install()
#from twisted.internet import reactor

import os
import sys
import traceback
import re
import glob
import datetime
import pygtk
pygtk.require('2.0')
import gtk
import apt.progress
import apt.progress.gtk2
import apt_pkg
import subprocess
import logging

from ui import JoliFrame

# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/desktop-recommends.remove
# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/desktop.remove
# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/standard-recommends.remove
# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/standard.remove
# https://git.jolicloud.org/gitweb/?p=system/build.git;a=blob;f=config-robby-lucid/packages.remove
BLACKLIST = [
    re.compile(r'(^firefox|^ubuntuone|^gwibber|^plymouth-theme-ubuntu|^cups|^foomatic|^ghostscript|^empathy|^linux.*generic$|^linux-headers-2.6.32|^python-ubuntuone|^xulrunner)'),
    
    'ubuntu-docs',
    'evolution',
    'genisoimage',
    'gnome-themes-ubuntu',
    'gucharmap',
    'popularity-contest',
    'pitivi',
    'brasero',
    'software-center',
    'system-config-printer-gnome',
    'ubiquity',
    'oem-config',
    'update-notifier',
    #'yelp', # This one must be installed (gnome-terminal, gnome-panel)

    'apparmor',
    'cups',
    'cups-bsd',
    'cups-client',
    'dc',
    'evince',
    'foomatic-db',
    'foomatic-db-engine',
    'foomatic-filters',
    'gdm',
    'genisoimage',
    'ghostscript-x',
    #'gnome-icon-theme', # This one must be installed (humanity-icon-theme, jolicloud-icon-theme)
    'gnome-spell',
    'gnome-themes-ubuntu',
    'gnome-themes-selected',
    'gucharmap',
    'nautilus-cd-burner',
    'openprinting-ppds',
    'pnm2ppa',
    'software-center', # exists ??
    'software-properties-gtk',
    'system-config-printer-gnome',
    'ubuntu-artwork',
    'ubuntu-sounds',
    'update-notifier',
    #'yelp', # This one must be installed (gnome-terminal, gnome-panel)

    'apparmor-utils',
    'manpages',
    'plymouth-theme-ubuntu-text', # exists ??
    'w3m',
    'popularity-contest',
]

INSTALL_TIME = 80 # 80 minutes

def update_ui():
    while gtk.events_pending():
        gtk.main_iteration()

def format_time(minutes):
    s = str(datetime.timedelta(minutes=minutes)).split(':')
    if s[0] == '0':
        if s[1] == '00':
            return 'Finishing...'
        return '%s minutes remaining...' % s[1].lstrip('0')
    else:
        return '%sh%s remaining...' % (s[0].lstrip('0'), s[1])

def set_default_login_manager(name='slim', path='/usr/bin/slim'):
    #echo "slim shared/default-x-display-manager select slim" | debconf-set-selections
    subprocess.Popen(
        'echo "%s shared/default-x-display-manager select %s" | debconf-set-selections' % (name, name),
        shell=True,
        stdout=subprocess.PIPE
    ).communicate()[0]
    
    #echo "slim shared/default-x-display-manager seen true" | debconf-set-selections
    subprocess.Popen(
        'echo "%s shared/default-x-display-manager seen true" | debconf-set-selections' % name,
        shell=True,
        stdout=subprocess.PIPE
    ).communicate()[0]
    
    #echo "slim slim/daemon_name string /usr/bin/slim" | debconf-set-selections
    subprocess.Popen(
        'echo "%s %s/daemon_name string %s" | debconf-set-selections' % (name, name, path),
        shell=True,
        stdout=subprocess.PIPE
    ).communicate()[0]
    
    #echo "/usr/bin/slim" > /etc/X11/default-display-manager
    f = open('/etc/X11/default-display-manager', 'w')
    f.write('%s\n' % path)
    f.close()

class JoliAptProgress(apt.progress.gtk2.GtkAptProgress):
    def _on_status_changed(self, progress, status, percent):
        """Called when the status changed."""
        while gtk.events_pending():
            gtk.main_iteration()

class JolicloudUpgrader(object):
    
    _win_bg_color = gtk.gdk.color_parse('#27282E')
    _box_bg_color = gtk.gdk.color_parse('#515664')
    _color_white = gtk.gdk.color_parse('white')
    
    _doing_upgrade = False
    _keys = ''
    
    def __init__(self):
        builder = gtk.Builder()
        builder.add_from_file(os.path.join(os.environ['JU_DATA_PATH'], 'upgrader.ui'))
        builder.connect_signals({
            'on_window_destroy' : gtk.main_quit,
        })
        #builder.connect_signals({
        #    'on_window_destroy' : reactor.stop,
        #})
        self.window = builder.get_object('window')
        self.window.connect('key_press_event', self.on_key_press)
        self.window.connect('map-event', self.on_map_event)
        self.window.modify_bg(gtk.STATE_NORMAL, self._win_bg_color)
        
        self.hbox = builder.get_object('hbox')
        
        self.main_frame = JoliFrame(bgcolor=self._win_bg_color, rounded=10)
        self.main_frame.set_size_request(600, 200)
        self.hbox.pack_start(self.main_frame, expand=False, fill=False, padding=20)
        self.hbox.reorder_child(self.main_frame, 1)
        box_builder = gtk.Builder()
        box_builder.add_from_file(os.path.join(os.environ['JU_DATA_PATH'], 'main_box.ui'))
        
        self.main_box = box_builder.get_object('main_box')
        self.progress = JoliAptProgress()
        self.progress._terminal.set_size_request(600, 250)
        self.progress.hide_terminal()
        # self.progress.show_terminal(True)
        self.progress.set_border_width(20)
        self.progress.open.connect('status-changed', self.open_status_changed)
        self.progress.fetch.connect('status-changed', self.fetch_status_changed)
        self.progress.install.connect('status-changed', self.install_status_changed)
        self.progress.install.connect('status-finished', self.install_status_finished)
        self.progress.install.connect('status-error', self.install_status_error)
        self.progress._label.hide()
        self.progress._label.modify_fg(gtk.STATE_NORMAL, self._color_white)
        
        self.main_box.pack_start(self.progress, expand=False, fill=False, padding=0)
        self.main_box.reorder_child(self.progress, 2)
        
        self.progress.show()
        
        self.main_frame.add(self.main_box)
        self.main_frame.show()
        
        self.window.fullscreen()
        self.window.stick()
        self.window.set_keep_above(True)
        self.window.show()
    
    def on_map_event(self, event, m):
        if self._doing_upgrade == False:
            self._doing_upgrade = True
            os.environ['RELEASE_UPGRADE_IN_PROGRESS'] = '1'
            os.environ['PYCENTRAL_NO_DPKG_QUERY'] = '1'
            os.environ['PYCENTRAL_FORCE_OVERWRITE'] = '1'
            os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
            cache = apt.cache.Cache(self.progress.open)
            cache.update()
            cache.open(None)
            cache.upgrade(True)
            def clean_packages():
                t = datetime.datetime.now()
                for package in cache:
                    update_ui()
                    for r in BLACKLIST:
                        if not isinstance(r, basestring) and r.match(package.name) and (package.markedInstall or package.markedUpgrade):
                            package.markDelete(purge=True) # mark_delete, autoFix=False ?
                        update_ui()
                for s in BLACKLIST:
                    if isinstance(s, basestring) and s in cache and (package.markedInstall or package.markedUpgrade):
                        cache[s].markDelete(purge=True) # mark_delete, autoFix=False ?
                    update_ui()
                logging.debug('time = %s' % (datetime.datetime.now() - t))
            if 'JU_DEBUG' not in os.environ:
                with cache.actiongroup():
                    clean_packages()
            else:
                print '[DEBUG] BLACKLIST'
            
            logging.debug('Packages that are going to be changed:')
            changed = apt.cache.FilteredCache(cache)
            changed.setFilter(apt.cache.MarkedChangesFilter()) # set_filter
            for p in changed:
                logging.debug('%s %d%d%d%d%d%d %d' % (
                    p.name,
                    p.markedDelete, # mark_delete
                    p.markedDowngrade, # mark_downgrade
                    p.markedInstall, # mark_install
                    p.markedKeep, # mark_keep
                    p.markedReinstall, # mark_reinstall
                    p.markedUpgrade, # mark_upgrade
                    p.candidate.size if p.candidate else 0
                ))
            logging.debug('---------')
            logging.debug('Required Download Size: %d' % cache.requiredDownload) # required_download
            
            cache.commit(self.progress.fetch, self.progress.install)
    
    def on_key_press(self, window, event):
        self._keys += event.string
        if self._keys.endswith('show'):
            self._keys = ''
            self.progress.show_terminal(True)
        elif self._keys.endswith('hide'):
            self._keys = ''
            self.progress.hide_terminal()
            
    
    def open_status_changed(self, event, status, percent):
        logging.debug('[OPEN] %s %d' % (status, percent))
        self.progress._label.set_text('Getting upgrade information...')
        if status in ('Reading package lists', 'Building dependency tree'):
            percent = 0
        if percent is None or percent == -1:
            self.progress._progressbar.pulse()
        else:
            self.progress._progressbar.set_fraction(percent/100.0*0.05)
        
    def fetch_status_changed(self, event, status, percent):
        logging.debug('[FETCH] %s %d' % (status, percent))
        if event.currentCPS != 0.0: # current_cps
            remaining_time = INSTALL_TIME + (event.totalBytes / event.currentCPS / 60) # total_bytes, current_cps
            self.progress._label.set_text('Downloading upgrade - %s ' % format_time(remaining_time))
        else:
            self.progress._label.set_text('Downloading upgrade...')
        if percent is None or percent == -1:
            self.progress._progressbar.pulse()
        else:
            self.progress._progressbar.set_fraction(0.05 + percent/100.0*0.2)
    
    def install_status_changed(self, event, status, percent):
        logging.debug('[INSTALL] %s %d' % (status, percent))
        remaining_time = INSTALL_TIME - (INSTALL_TIME * percent / 100)
        self.progress._label.set_text('Installing upgrade - %s' % format_time(remaining_time))
        if percent is None or percent == -1:
            self.progress._progressbar.pulse()
        else:
            self.progress._progressbar.set_fraction(0.25 + percent/100.0*0.75)
    
    def install_status_finished(self, event):        
        self.progress._progressbar.set_fraction(1)
        self.progress._label.set_text('Finished!')
        
        try:
            # dir_cache = apt_pkg.config.find_dir("Dir::Cache::archives") # This doesn't work after an upgrade
            # Assuming the cahce dir is at the default place:
            dir_cache = '/var/cache/apt/archives/'
            for deb in glob.glob(dir_cache + '*.deb') + glob.glob(dir_cache + 'partial/*.deb'):
               os.unlink(deb)
        except Exception:
            traceback.print_exc(file=sys.stdout)
        if 'JU_DEBUG' not in os.environ:
            self.progress._label.set_text('Your system is going to reboot...')
            os.system('/sbin/reboot')
        else:
            logging.debug("[DEBUG] os.system('/sbin/reboot')")
    
    def install_status_error(self, param):
        logging.debug('[ERROR] %s' % param)

def do_upgrade():
    logging.basicConfig(
        filename=os.environ['JU_LOG_PATH'],
        level=logging.DEBUG,
        format='[%(asctime)-15s] %(message)s'
    )
    try:
        set_default_login_manager('slim', '/usr/bin/slim')
        if 'JU_DEBUG' not in os.environ:
            for bin in glob.glob('/var/cache/apt/*.bin'):
                os.unlink(bin)
        app = JolicloudUpgrader()
        gtk.main()
    except Exception:
        traceback.print_exc(file=sys.stdout)
        set_default_login_manager('gdm', '/usr/sbin/gdm')

if __name__ == '__main__':
    do_upgrade()

