#!/usr/bin/env python

import gtk
import gobject
import cairo

def rounded_rectangle(context, corner_radius, x, y, width, height):
    corner_radius = min(corner_radius, min(width, height) / 2)
    x2, y2 = x + width, y + height
    half_corner = corner_radius / 2
    
    context.move_to(x + corner_radius, y)
    context.line_to(x2 - corner_radius, y)
    context.curve_to(x2 - half_corner, y, x2, y + half_corner, x2, y + corner_radius)
    context.line_to(x2, y2 - corner_radius)
    context.curve_to(x2, y2 - half_corner, x2 - half_corner, y2, x2 - corner_radius, y2)
    context.line_to(x + corner_radius, y2)
    context.curve_to(x + half_corner, y2, x, y2 - half_corner, x, y2 - corner_radius)
    context.line_to(x, y + corner_radius)
    context.curve_to(x, y + half_corner, x + half_corner, y, x + corner_radius, y)

def color_to_cairo_rgba(c, alpha=1):
    return c.red/65535.0, c.green/65535.0, c.blue/65535.0, alpha

class JoliFrame(gtk.Frame):
    def __init__(self, bgcolor='white', transparency=1, rounded=10):
        self.set_bgcolor(bgcolor)
        self.transparency = transparency
        self.rounded = rounded
        gtk.Frame.__init__(self)

    def set_bgcolor(self, color):
        if isinstance(color, gtk.gdk.Color): self.bgcolor = color
        elif isinstance(color, str): self.bgcolor = gtk.gdk.color_parse(color)

    def do_expose_event(self, event):
        self.set_shadow_type(gtk.SHADOW_NONE)
        x, y, w, h = self.allocation
        
        c = self.window.cairo_create()
        c.set_source_rgba(*color_to_cairo_rgba(self.bgcolor))
        rounded_rectangle(c, self.rounded, *self.allocation)
        c.fill()
        gtk.Frame.do_expose_event(self, event)
    
gobject.type_register(JoliFrame)
