#!/usr/bin/env python

#from twisted.internet import gtk2reactor
#gtk2reactor.install()
#from twisted.internet import reactor

import os
import sys
import traceback
import re
import glob
import datetime
import string
import pygtk
pygtk.require('2.0')
import gtk
import apt.progress
import apt.progress.gtk2
import apt_pkg
import subprocess
import logging
import apt.debfile

from ui import JoliFrame

# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/desktop-recommends.remove
# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/desktop.remove
# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/standard-recommends.remove
# https://git.jolicloud.org/gitweb/?p=system/jolicloud-meta.git;a=blob;f=jolicloud/standard.remove
# https://git.jolicloud.org/gitweb/?p=system/build.git;a=blob;f=config-robby-lucid/packages.remove
BLACKLIST = [
    re.compile(r'(^firefox|^ubuntuone|^gwibber|^plymouth-theme-ubuntu|^empathy|^linux.*generic$|^linux-headers-2.6.32|^python-ubuntuone|^xulrunner)'),
    
    re.compile(r'(^adium|^aisleriot$|^app-install|^bcmwl|^branding-ubuntu|^computer-jantior|^erlang|^gnome-doc-utils$|^gnome-games^gnome-mahjongg$|^gnome-sudoku$|^groff|^empathy|^espeak|^evolution$|^firefox|^jockey|^gdm|^gucharmap$|^libqt|^gnome-bluetooth$|^linux.*generic$|^linux-headers-2.6.32|^manpages|^openoffice.org|^plymouth-theme-ubuntu|^python-ubuntuone|^quadrapassel$|^rhythmbox|^software-|^telepathy|^tomboy|^tsclient|^ubuntuone|^update-notifier$|^uno|^w3m|^xulrunner)'),
    
    'example-content',
    'evolution-common',
    'evolution-data-server',
    'evolution-webcal',
    'freepats',
    'gbrainy',
    'gdb',
    'gnomine',
    'genisoimage',
    'gnome-accessibility-themes',
    'gnome-themes-ubuntu',
    'gnome-mag',
    'libsane',
    'light-themes',
    'man-db',
    'netbook-launcher',
    'pitivi',
    'popularity-contest',
    'system-config-printer-gnome',
    'ubuntu-docs',
    'ubuntu-mono',
    'ubuntu-wallpapers',
    #'yelp', # This one must be installed (gnome-terminal, gnome-panel)
    
    'ubuntu-docs',
    'evolution',
    'genisoimage',
    'gnome-themes-ubuntu',
    'gucharmap',
    'popularity-contest',
    'pitivi',
    'software-center',
    'system-config-printer-gnome',
    'ubiquity',
    'oem-config',
    'update-notifier',
    #'yelp', # This one must be installed (gnome-terminal, gnome-panel)

    'apparmor',
    'dc',
    'evince',
    'gdm',
    'genisoimage',
    #'gnome-icon-theme', # This one must be installed (humanity-icon-theme, jolicloud-icon-theme)
    'gnome-spell',
    'gnome-themes-ubuntu',
    'gnome-themes-selected',
    'gucharmap',
    'nautilus-cd-burner',
    'software-center', # exists ??
    'software-properties-gtk',
    'system-config-printer-gnome',
    'ubuntu-artwork',
    'ubuntu-sounds',
    'update-notifier',
    #'yelp', # This one must be installed (gnome-terminal, gnome-panel)

    'apparmor-utils',
    'manpages',
    'plymouth-theme-ubuntu-text', # exists ??
    'w3m',
    'popularity-contest',
]

DEB_JOLICLOUD_APT_CONFIG = '/var/lib/jolicloud-upgrader/archive/jolicloud-apt-config_1.1.8_all.deb'
DEB_JOLICLOUD_APT_REPOSITORY = '/var/lib/jolicloud-upgrader/archive/jolicloud-apt-repository_1.1.8_all.deb'

INSTALL_TIME = 80 # 80 minutes

def update_ui():
    while gtk.events_pending():
        gtk.main_iteration()

def format_time(minutes):
    s = str(datetime.timedelta(minutes=minutes)).split(':')
    if s[0] == '0':
        if s[1] == '00':
            return 'Finishing up...'
        if s[1] == '01':
            return 'approximately %s minute remaining...' % s[1].lstrip('0')
        else:
            return 'approximately %s minutes remaining...' % s[1].lstrip('0')
    else:
        return 'approximately %s hour %s minutes remaining...' % (s[0].lstrip('0'), s[1])

def set_default_login_manager(name='slim', path='/usr/bin/slim'):
    #echo "slim shared/default-x-display-manager select slim" | debconf-set-selections
    subprocess.Popen(
        'echo "%s shared/default-x-display-manager select %s" | debconf-set-selections' % (name, name),
        shell=True,
        stdout=subprocess.PIPE
    ).communicate()[0]
    
    #echo "slim shared/default-x-display-manager seen true" | debconf-set-selections
    subprocess.Popen(
        'echo "%s shared/default-x-display-manager seen true" | debconf-set-selections' % name,
        shell=True,
        stdout=subprocess.PIPE
    ).communicate()[0]
    
    #echo "slim slim/daemon_name string /usr/bin/slim" | debconf-set-selections
    subprocess.Popen(
        'echo "%s %s/daemon_name string %s" | debconf-set-selections' % (name, name, path),
        shell=True,
        stdout=subprocess.PIPE
    ).communicate()[0]
    
    #echo "/usr/bin/slim" > /etc/X11/default-display-manager
    f = open('/etc/X11/default-display-manager', 'w')
    f.write('%s\n' % path)
    f.close()

def kill_screensaver():
    try:
        subprocess.Popen(
            "killall gnome-screensaver",
            shell=True,
            stdout=subprocess.PIPE
        ).communicate()[0];
    except Exception:
        pass

def setup_new_repos():
    os.environ['RELEASE_UPGRADE_IN_PROGRESS'] = '1'
    os.environ['PYCENTRAL_NO_DPKG_QUERY'] = '1'
    os.environ['PYCENTRAL_FORCE_OVERWRITE'] = '1'
    os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
    
    # Install new apt repos
    subprocess.Popen(
        ['/usr/bin/dpkg', '-i', '--force-confnew', DEB_JOLICLOUD_APT_CONFIG, DEB_JOLICLOUD_APT_REPOSITORY],
        stdout=subprocess.PIPE
    ).communicate()[0]

def is_atom_cpu():
    p = open('/proc/cpuinfo')
    cpu_family = None
    cpu_model = None
    for line in p.readlines():
        try:
            (key, value) = line.strip().split(':')
            if key.strip() == 'cpu family':
                cpu_family = string.atoi(value.strip())
            if key.strip() == 'model':
                cpu_model = string.atoi(value.strip())
            if cpu_family != None and cpu_model != None:
                break
        except (ValueError, IndexError):
            pass
    p.close()

    if cpu_family == 6 and cpu_model == 28: # Atom CPU
        return True
    return False

class JoliAptProgress(apt.progress.gtk2.GtkAptProgress):
    def _on_status_changed(self, progress, status, percent):
        """Called when the status changed."""
        while gtk.events_pending():
            gtk.main_iteration()

class JolicloudUpgrader(object):
    
    _win_bg_color = gtk.gdk.color_parse('#27282E')
    _box_bg_color = gtk.gdk.color_parse('#515664')
    _color_white = gtk.gdk.color_parse('white')
    
    _doing_upgrade = False
    _slim_is_default_login_manager = False
    _keys = ''
    
    def __init__(self):
        builder = gtk.Builder()
        builder.add_from_file(os.path.join(os.environ['JU_DATA_PATH'], 'upgrader.ui'))
        builder.connect_signals({
            'on_window_destroy' : gtk.main_quit,
        })
        #builder.connect_signals({
        #    'on_window_destroy' : reactor.stop,
        #})
        self.window = builder.get_object('window')
        self.window.connect('key_press_event', self.on_key_press)
        self.window.connect('map-event', self.on_map_event)
        self.window.modify_bg(gtk.STATE_NORMAL, self._win_bg_color)
        
        self.hbox = builder.get_object('hbox')
        
        self.main_frame = JoliFrame(bgcolor=self._win_bg_color, rounded=10)
        self.main_frame.set_size_request(600, 200)
        self.hbox.pack_start(self.main_frame, expand=False, fill=False, padding=20)
        self.hbox.reorder_child(self.main_frame, 1)
        box_builder = gtk.Builder()
        box_builder.add_from_file(os.path.join(os.environ['JU_DATA_PATH'], 'main_box.ui'))
        
        self.main_box = box_builder.get_object('main_box')
        self.progress = JoliAptProgress()
        self.progress._terminal.set_size_request(600, 250)
        if 'JU_DEBUG' not in os.environ:
            self.progress.hide_terminal()
        else:
            self.progress.show_terminal(True)
        self.progress.set_border_width(20)
        self.progress.open.connect('status-changed', self.open_status_changed)
        self.progress.fetch.connect('status-changed', self.fetch_status_changed)
        self.progress.fetch.connect('status-finished', self.fetch_status_finished)
        self.progress.install.connect('status-changed', self.install_status_changed)
        self.progress.install.connect('status-finished', self.install_status_finished)
        self.progress.install.connect('status-error', self.install_status_error)
        self.progress.install.connect('status-conffile', self.install_status_conffile)
        self.progress._label.hide()
        self.progress._label.modify_fg(gtk.STATE_NORMAL, self._color_white)
        
        self.main_box.pack_start(self.progress, expand=False, fill=False, padding=0)
        self.main_box.reorder_child(self.progress, 2)
        
        self.progress.show()
        
        self.main_frame.add(self.main_box)
        self.main_frame.show()
        
        if 'JU_DEBUG' not in os.environ:
            self.window.fullscreen()
            self.window.stick()
            self.window.set_keep_above(True)
        self.window.show()
    
    def on_map_event(self, event, m):
        if self._doing_upgrade == False:
            self._doing_upgrade = True
            
            apt_pkg.InitConfig()
            apt_pkg.Config.Set('DPkg::Options::', '--force-overwrite')
            apt_pkg.Config.Set('DPkg::Options::', '--force-confnew')
            apt_pkg.InitSystem()
            cache = apt.cache.Cache(self.progress.open)
            apt_pkg.InitConfig()
            apt_pkg.Config.Set('DPkg::Options::', '--force-overwrite')
            apt_pkg.Config.Set('DPkg::Options::', '--force-confnew')
            apt_pkg.InitSystem()
            # It seems that apt.cache redefine those env variable
            os.environ['RELEASE_UPGRADE_IN_PROGRESS'] = '1'
            os.environ['PYCENTRAL_NO_DPKG_QUERY'] = '1'
            os.environ['PYCENTRAL_FORCE_OVERWRITE'] = '1'
            os.environ['DEBIAN_FRONTEND'] = 'noninteractive'
            cache.update()
            cache.open(None)
            cache.upgrade(True)
            
            def clean_packages():
                t = datetime.datetime.now()
                for package in cache:
                    update_ui()
                    for r in BLACKLIST:
                        if not isinstance(r, basestring) and r.match(package.name) and (package.markedInstall or package.markedUpgrade):
                            package.markDelete(purge=True) # mark_delete, autoFix=False ?
                            logging.debug('PACKAGE TO BE DELETED: %s' % package)
                        update_ui()
                for s in BLACKLIST:
                    if isinstance(s, basestring) and s in cache and (package.markedInstall or package.markedUpgrade):
                        cache[s].markDelete(purge=True) # mark_delete, autoFix=False ?
                        logging.debug('PACKAGE TO BE DELETED: %s' % cache[s])
                    update_ui()
                logging.debug('time = %s' % (datetime.datetime.now() - t))
            with cache.actiongroup():
                if is_atom_cpu():
                    logging.debug('Atom CPU detected, installing linux-jolicloud-atom kernel.')
                    cache['linux-jolicloud-atom'].markInstall()
                    cache['linux-headers-jolicloud-atom'].markInstall()
                    cache['linux-jolicloud'].markDelete(purge=True)
                    cache['linux-headers-jolicloud'].markDelete(purge=True)
                clean_packages()
            
            logging.debug('Packages that are going to be changed:')
            changed = apt.cache.FilteredCache(cache)
            changed.setFilter(apt.cache.MarkedChangesFilter()) # set_filter
            for p in changed:
                logging.debug('%s %d%d%d%d%d%d %d' % (
                    p.name,
                    p.markedDelete, # mark_delete
                    p.markedDowngrade, # mark_downgrade
                    p.markedInstall, # mark_install
                    p.markedKeep, # mark_keep
                    p.markedReinstall, # mark_reinstall
                    p.markedUpgrade, # mark_upgrade
                    p.candidate.size if p.candidate else 0
                ))
            logging.debug('---------')
            logging.debug('Required Download Size: %d' % cache.requiredDownload) # required_download
            
            apt_pkg.InitConfig()
            apt_pkg.Config.Set('DPkg::Options::', '--force-overwrite')
            apt_pkg.Config.Set('DPkg::Options::', '--force-confnew')
            apt_pkg.InitSystem()
            cache.commit(self.progress.fetch, self.progress.install)
    
    def on_key_press(self, window, event):
        self._keys += event.string
        if self._keys.endswith('show'):
            self._keys = ''
            self.progress.show_terminal(True)
        elif self._keys.endswith('hide'):
            self._keys = ''
            self.progress.hide_terminal()
    
    def open_status_changed(self, event, status, percent):
        logging.debug('[OPEN] %s %d' % (status, percent))
        self.progress._label.set_text('Getting upgrade information...')
        if status in ('Reading package lists', 'Building dependency tree'):
            percent = 0
        if percent is None or percent == -1:
            self.progress._progressbar.pulse()
        else:
            self.progress._progressbar.set_fraction(percent/100.0*0.05)
        
    def fetch_status_changed(self, event, status, percent):
        logging.debug('[FETCH] %s %d' % (status, percent))
        if event.currentCPS != 0.0: # current_cps
            remaining_time = INSTALL_TIME + (event.totalBytes / event.currentCPS / 60) # total_bytes, current_cps
            self.progress._label.set_text('Downloading upgrade - %d%% complete - %s ' % ( percent, format_time(remaining_time)))
        else:
            self.progress._label.set_text('Downloading upgrade...')
        if percent is None or percent == -1:
            self.progress._progressbar.pulse()
        else:
            self.progress._progressbar.set_fraction(0.05 + percent/100.0*0.2)
    
    def fetch_status_finished(self, progress):
        logging.debug('[FETCH FINISHED] %s' % progress)
        if progress.currentItems < progress.totalItems:
            logging.debug('[FETCH FAILED] Stoping the upgrader and restarting nickel-browser')
            subprocess.Popen(
                'killall nickel-browser ; killall nickel-browser',
                shell=True,
                stdout=subprocess.PIPE
            ).communicate()[0];
            gtk.main_quit()
        else:
            if self._slim_is_default_login_manager == False:
                set_default_login_manager('slim', '/usr/bin/slim')
                self._slim_is_default_login_manager = True
    
    def install_status_changed(self, event, status, percent):
        logging.debug('[INSTALL] %s %d' % (status, percent))
        remaining_time = INSTALL_TIME - (INSTALL_TIME * percent / 100)
        self.progress._label.set_text('Installing upgrade - %s' % format_time(remaining_time))
        if percent is None or percent == -1:
            self.progress._progressbar.pulse()
        else:
            self.progress._progressbar.set_fraction(0.25 + percent/100.0*0.75)
    
    def install_status_finished(self, event):
        logging.debug('[INSTALL FINISHED] %s' % event)
        self.progress._progressbar.set_fraction(1)
        self.progress._label.set_text('Finished!')
        
        try:
            if 'JU_DEBUG' not in os.environ:
                # dir_cache = apt_pkg.config.find_dir("Dir::Cache::archives") # This doesn't work after an upgrade
                # Assuming the cahce dir is at the default place:
                dir_cache = '/var/cache/apt/archives/'
                for deb in glob.glob(dir_cache + '*.deb') + glob.glob(dir_cache + 'partial/*.deb'):
                   os.unlink(deb)
        except Exception:
            traceback.print_exc(file=sys.stdout)

        # Ensure the symbolic links /initrd.img and /vmlinuz point to the
        # correct kernel. First, identify the correct initrd and vmlinuz
        # images.
        initrd_image = subprocess.Popen(
            "ls -1 /boot/initrd* | tail -1",
            shell=True,
            stdout=subprocess.PIPE
        ).communicate()[0].rstrip();
        vmlinuz_image = subprocess.Popen(
            "ls -1 /boot/vmlinuz* | tail -1",
            shell=True,
            stdout=subprocess.PIPE
        ).communicate()[0].rstrip();

        # Moving the current symlink to the .old file, recreate the symlink.
        if os.path.exists( '/initrd.img.old' ):
            os.unlink( '/initrd.img.old' )
        if os.path.exists( '/initrd.img' ):
            os.rename( '/initrd.img', '/initrd.img.old' );
        os.symlink( initrd_image, '/initrd.img' )
        if os.path.exists( '/vmlinuz.old' ):
            os.unlink( '/vmlinuz.old' )
        if os.path.exists( '/vmlinuz' ):
            os.rename( '/vmlinuz', '/vmlinuz.old' );
        os.symlink( vmlinuz_image, '/vmlinuz' )

        # Manual hacks for hardware-specific devices that are
        # inconsequential to other hardware. Specifically, ensure nouveau
        # is disabled, and an outdated libdrm-poulsbo.so are not
        # referenced:

        # Hack #1: Remove depricated pouslbo DRM library.
        #
        # Remove /usr/lib/libdrm-poulsbo.so.2 from /etc/ld.so.preload if found.
        # This was a file used under Jolicloud 1.0 for Poulsbo users but no
        # longer exists for Jolicloud 1.1.
        # FIXME: Only remove libdrm-poulsbo.so.2 from this preload file,
        # but keep everything else, just incase the user modified this file
        # on their own.
        if os.path.exists( '/etc/ld.so.preload' ):
            os.unlink( '/etc/ld.so.preload' );

        # Hack #2: Inhibit nouveau driver for nvidia.
        #
        # Create /etc/modprobe.d/nvidia-graphics-drivers.conf and blacklist
        # the "nouveau" module. This automatically gives precedence to the
        # working "nvidia" driver, but keeps nouveau available for those
        # who wish to test it.
        f = open( '/etc/modprobe.d/nvidia-graphics-drivers.conf', 'w' )
        f.write( 'blacklist nouveau\n' );
        f.close()

        # Update the initrd just to be super-safe that we've regenerated an
        # initrd for the correct kernel. It is vital that the preceding
        # hardware-specific hacks make it into the correct initrd!
        subprocess.Popen(
            "update-initramfs -u",
            shell=True,
            stdout=subprocess.PIPE
        ).communicate()[0];

        if 'JU_DEBUG' not in os.environ:
            self.progress._label.set_text('Your system is going to reboot...')
            os.system('/sbin/reboot')
        else:
            logging.debug("[DEBUG] os.system('/sbin/reboot')")
    
    def install_status_error(self, error):
        logging.debug('[ERROR] %s' % error)
        try:
            self.progress._label.set_text('Installing upgrade...')
            self.progress._progressbar.pulse()
            subprocess.Popen(
                'apt-get -f -y install',
                shell=True,
                stdout=subprocess.PIPE
            ).communicate()[0];
        except Exception:
            logging.debug('[ERROR] Exception')
    
    def install_status_conffile(self, current, new):
        logging.debug('[CONFFILE] Warning, conffile prompt, using new version.')

def do_upgrade():
    logging.basicConfig(
        filename=os.environ.get('JU_LOG_PATH', 'jolicloud-upgrader.log'),
        level=logging.DEBUG,
        format='[%(asctime)-15s] %(message)s'
    )
    jolicloud_version = subprocess.Popen(
        'lsb_release -ds',
        shell=True,
        stdout=subprocess.PIPE
    ).communicate()[0].strip();
    if jolicloud_version != 'Jolicloud robby (v1.0 Release)':
        logging.debug('Not running on Jolicloud 1.0, exiting.')
        return
    try:
        if 'JU_DEBUG' not in os.environ:
            for bin in glob.glob('/var/cache/apt/*.bin'):
                os.unlink(bin)
        kill_screensaver()
        setup_new_repos()
        app = JolicloudUpgrader()
        gtk.main()
    except Exception:
        traceback.print_exc(file=sys.stdout)
        if self._slim_is_default_login_manager == True:
            set_default_login_manager('gdm', '/usr/sbin/gdm')

if __name__ == '__main__':
    do_upgrade()
